''***********************************
''*  VGA High-Res Text Driver v1.1  *
''*  (C) 2006 Parallax, Inc.        *
''*  Modified by Michael Green      *
''*  to use 640 x 480 default and   *
''*  to use a pointer to workspace  *
''*  Application interface portion  *
''***********************************
''
'' This object generates a 1024x768 VGA signal which contains 128 columns x 64
'' rows of 8x12 characters. Each row can have a unique forground/background
'' color combination and each character can be inversed. There are also two
'' cursors which can be independently controlled (ie. mouse and keyboard). A
'' sync indicator signals each time the screen is refreshed (you may ignore).
''
'' You must provide buffers for the screen, colors, cursors, and sync. Once
'' started, all interfacing is done via memory. To this object, all buffers are
'' read-only, with the exception of the sync indicator which gets written with
'' -1. You may freely write all buffers to affect screen appearance. Have fun!
''

CON

{
' 1024 x 768 @ 57Hz settings: 128 x 64 characters

  hp = 1024     'horizontal pixels
  vp = 768      'vertical pixels
  hf = 16       'horizontal front porch pixels
  hs = 96       'horizontal sync pixels
  hb = 176      'horizontal back porch pixels
  vf = 1        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 60       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}
{
' 800 x 600 @ 75Hz settings: 100 x 50 characters

  hp = 800      'horizontal pixels
  vp = 600      'vertical pixels
  hf = 40       'horizontal front porch pixels
  hs = 128      'horizontal sync pixels
  hb = 88       'horizontal back porch pixels
  vf = 1        'vertical front porch lines
  vs = 4        'vertical sync lines
  vb = 23       'vertical back porch lines
  hn = 0        'horizontal normal sync state (0|1)
  vn = 0        'vertical normal sync state (0|1)
  pr = 50       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}
'{
' 640 x 480 @ 69Hz settings: 80 x 40 characters

  hp = 640      'horizontal pixels
  vp = 480      'vertical pixels
  hf = 24       'horizontal front porch pixels
  hs = 40       'horizontal sync pixels
  hb = 128      'horizontal back porch pixels
  vf = 9        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 30       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
'}

' columns and rows

  cols = hp / 8
  rows = vp / 12

OBJ
  ldr : "OS_loader"
  
''   screenPtr = Pointer to 8,192 bytes containing ASCII codes for each of the
''               128x64 screen characters. Each byte's top bit controls color
''               inversion while the lower seven bits provide the ASCII code.
''               Screen memory is arranged left-to-right, top-to-bottom.
''
''               screen byte example: %1_1000001 = inverse "A"
''
''    colorPtr = Pointer to 64 words which define the foreground and background
''               colors for each row. The lower byte of each word contains the
''               foreground RGB data for that row, while the upper byte
''               contains the background RGB data. The RGB data in each byte is
''               arranged as %RRGGBB00 (4 levels each).
''
''               color word example: %%0020_3300 = gold on blue
''
''   cursorPtr = Pointer to 6 bytes which control the cursors:
''
''               bytes 0,1,2: X, Y, and MODE of cursor 0
''               bytes 3,4,5: X, Y, and MODE of cursor 1
''
''               X and Y are in terms of screen characters
''               (left-to-right, top-to-bottom)
''
''               MODE uses three bottom bits:
''
''                      %x00 = cursor off
''                      %x01 = cursor on
''                      %x10 = cursor on, blink slow
''                      %x11 = cursor on, blink fast
''                      %0xx = cursor is solid block
''                      %1xx = cursor is underscore
''
''               cursor example: 127, 63, %010 = blinking block in lower-right
''
''    videoPtr = Pointer to work area for Hires VGA driver.  See below for specifics
CON
  vga_sync    =  0                             ' Sync.  Set to -1 at start of refresh
  vga_cog     =  4                             ' Video cog bitmask      
  vga_col     =  8                             ' Current column (screen relative)
  vga_row     =  9                             ' Current row (screen relative)
  vga_eol     = 10                             ' At end of line (for lazy return)
  vga_wleft   = 11                             ' Current window column offset
  vga_wtop    = 12                             ' Current window top offset
  vga_wcols   = 13                             ' Current window # cols
  vga_wrows   = 14                             ' Current window # rows
  vga_mode    = 15                             ' %abc_nnnn  a = insert mode
                                               '            b = update cursor 1 X,Y
                                               '            c = update cursor 0 X,Y
                                               '            n = current window number
  vga_pcnt    = 16                             ' # of parameters to collect
  vga_pptr    = 17                             ' Index of current parameter byte
  vga_flag    = 18                             ' Saved "command" character
  vga_params  = 19                             ' Buffer for parameters (5)
  vga_wsave   = 24                             ' Save area for window state (128)
  
  insertMask  = %100_0000                      ' Used to isolate insert mode in vga_mode
  updateMask  = %001_0000                      ' Used to isolate cursor update bits
  winInfoSize = vga_mode - vga_col + 1         ' # bytes used for window state

'' Control codes.  For convenience, supplying a # cols and # rows of zero will use the maximum
'' allowable values (screen width - x, screen height - y).  Because of the limitations on colors
'' (foreground and background for each row only), the Color command will set the colors for a
'' specific row.  The colors do not scroll with the text.

  ClrWin      = $00                            ' clear window  
  Home        = $01                            ' home in window
  MoveXY      = $02                            ' move to X,Y in window (X and Y follow)
  CursLt      = $03                            ' cursor left
  CursRt      = $04                            ' cursor right
  CursUp      = $05                            ' cursor up
  CursDn      = $06                            ' cursor down
  Bsp         = $08                            ' backspace
  Tab         = $09                            ' tab (8 spaces per)
  Lf          = $0A                            ' linefeed
  ClrEol      = $0B                            ' clear to end of line
  Cr          = $0D                            ' carriage return
  MoveX       = $0E                            ' move to X in window (X follows)
  MoveY       = $0F                            ' move to Y in window (Y follows)
  SetWin      = $10                            ' set bounds of window (1-15,x,y,cols,rows)
  UseWin      = $11                            ' change to selected window (0-15)
  Color       = $13                            ' change the colors of a row (Y, FG, BG follow)
  Insert      = $14                            ' insert (move line to right one column)
  Cursor      = $15                            ' set cursor (%nuxyy follows)
                                               ' n = cursor #, u = underline(1)/solid(0)
                                               ' yy = off(00)/on(01)/slow(10)/fast(11)
  ScrollUp    = $16                            ' scroll up the current window
  ScrollDn    = $17                            ' scroll down the current window
  EscChr      = $1F                            ' (ESC) print next character C as-is (C follows)
  Delete      = $7F                            ' delete current character
  Alt00       = $FF                            ' changed to $00 by out (for use in string())
  ClrScr      = Alt00                          ' equivalent to ClrWin (for use in string())

  absX        = %0000                          ' X value is relative to screen coordinates             
  relX        = %0001                          ' X value is relative to current values
  winX        = %0010                          ' X value is relative to window coordinates
  absY        = %0000                          ' Y value is relative to screen coordinates
  relY        = %0100                          ' Y value is relative to current values
  winY        = %1000                          ' Y value is relative to window coordinates

PUB out(c) | i, j, k, p, f, x, y, t
'' Output a character.  This routine decodes the character stream, interpreting control codes.
'' These mostly follow the codes used with other text drivers with the exception of color since
'' hires vga color is very limited and associated with an entire screen row.

  p := long[ldr#videoPtr]
  if c == $FF
    c~
  if f := byte[p+vga_flag]
    byte[p+vga_params][byte[p+vga_pptr]++] := c
    if byte[p+vga_pptr] == byte[p+vga_pcnt]
      case f
        MoveXY:
          updateCursor(byte[p+vga_params][0],c,winX|winY,false)
          byte[p+vga_eol]~
        SetWin:
          i := byte[p+vga_params][0]
          if i > 0 and i < 16
            i := i * winInfoSize + p + vga_wsave - vga_col
            x := byte[p+vga_params][1] <# cols - 1
            byte[i+vga_col] := byte[i+vga_wleft] := x
            y := byte[p+vga_params][2] <# rows - 1
            byte[i+vga_row] := byte[i+vga_wtop] := y            
            if (t := byte[p+vga_params][3] <# cols - x) == 0
              t := cols - x
            byte[i+vga_wcols] := t
            if (t := c <# rows - y) == 0
              t := rows - y
            byte[i+vga_wrows] := t
            byte[i+vga_eol]~
            byte[i+vga_mode]~
        UseWin:
          if c < 16
            i := (byte[p+vga_mode] & %1111) * winInfoSize + p + vga_wsave
            j := c * winInfoSize + p + vga_wsave
            bytemove(i,p+vga_col,winInfoSize)
            bytemove(p+vga_col,j,winInfoSize)
            byte[p+vga_mode] := (byte[p+vga_mode] & !%1111) | c
            updateCursor(0,0,relX|relY,false)
        Color:
          y := byte[p+vga_params][0] <# rows - 1
          word[long[ldr#colorPtr]][y] := (c << 8) | byte[p+vga_params][1]
        MoveX:
          updateCursor(c,0,winX|relY,false)
          byte[p+vga_eol]~
        MoveY:
          updateCursor(0,c,relX|winY,false)
        Cursor:
          i := c >> 4 & 1
          j := updateMask << i
          byte[long[ldr#cursorPtr+3*i]][2] := c & %111
          byte[p+vga_mode] := (byte[p+vga_mode] & !j) | ((c & %1000 > 0) & j)
          updateCursor(0,0,relX|relY,false)
        EscChr:
          print(c)
      byte[p+vga_flag]~
  else
    case c & $FF
      ClrWin:
        y := byte[p+vga_wtop]
        t := byte[p+vga_wrows]
        repeat i from y to y + t - 1
          fillwith(byte[p+vga_wleft],i,byte[p+vga_wcols]," ")
        updateCursor(0,0,winX|winY,false)
        byte[p+vga_eol]~
      Home:     
        updateCursor(0,0,winX|winY,false)
        byte[p+vga_eol]~
      MoveXY:
        byte[p+vga_flag] := c
        byte[p+vga_pptr]~
        byte[p+vga_pcnt] := 2
      CursLt:
        updateCursor(-1,0,relX|relY,false)
        byte[p+vga_eol]~
      CursRt:
        if byte[p+vga_eol]== 0
          updateCursor(+1,0,relX|relY,true)
      CursUp:
        updateCursor(0,-1,relX|relY,false)
      CursDn:
        updateCursor(0,+1,relX|relY,false)
      SetWin:
        byte[p+vga_flag] := c
        byte[p+vga_pptr]~
        byte[p+vga_pcnt] := 5
      Bsp:
        x := byte[p+vga_col] - byte[p+vga_wleft]
        y := byte[p+vga_row] - byte[p+vga_wtop]
        if x | y
          if x
            x--
          else
            x := byte[p+vga_wcols] - 1
            y--
          updateCursor(x,y,winX|winY,false)
          t := byte[p+vga_row] * cols + byte[p+vga_col]
          byte[long[ldr#screenPtr] + t] := " "
      Tab:
        repeat
          print(" ")
        while (byte[p+vga_col] - byte[p+vga_wleft]) & 7
      Lf:
        if (byte[p+vga_row] - byte[p+vga_wtop]) == byte[p+vga_wrows] - 1
          x := byte[p+vga_col]
          y := byte[p+vga_row]
          newline
          updateCursor(x,y,absX|absY,false)
        else
          updateCursor(0,+1,relX|relY,false)
      ClrEol:
        if byte[p+vga_eol] == 0
          cleartoeol
      Color:
        byte[p+vga_flag] := c
        byte[p+vga_pptr]~
        byte[p+vga_pcnt] := 3
      Cr:
        if byte[p+vga_eol] == 0
          cleartoeol
        newline
      MoveX, MoveY, UseWin, Cursor, EscChr:
        byte[p+vga_flag] := c
        byte[p+vga_pptr]~
        byte[p+vga_pcnt] := 1
      Insert:
        if byte[p+vga_eol] == 0
          i := long[ldr#screenPtr] + byte[p+vga_row] * cols
          j := byte[p+vga_col]
          k := byte[p+vga_wleft] + byte[p+vga_wcols] - 1
          repeat t from 0 to k - j
            byte[i + k - t] := byte[i + k - t - 1]
          byte[i + j] := " "
      Delete:
        if byte[p+vga_eol] == 0
          i := long[ldr#screenPtr] + byte[p+vga_row] * cols
          j := i + byte[p+vga_col]
          k := byte[p+vga_wleft] + byte[p+vga_wcols] - 1
          bytemove(j,j+1,k - byte[p+vga_col])
          byte[i+k] := " "
      ScrollUp:
        scrollUpWin
      ScrollDn:
        scrollDownWin
      other:
        print(c)

PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB decFixed(value,width) | i, p, f, a, b, c
'' Print a decimal number of fixed width
  p := @a
  f~
  if value < 0
    -value
    byte[p++] := "-"
  i := 1_000_000_000
  repeat 10
    if value => i
      byte[p++] := value / i + "0"
      value //= i
      f~~
    elseif f or i == 1
      byte[p++] := "0"
    i /= 10
  byte[p]~
  i := strsize(@a)
  if i < width
    repeat width - i
      out(" ")
  str(@a)
  
PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB screenChar(x, y) | p
'' Return the character at the absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return 0
  return byte[y * cols + x + long[ldr#screenPtr]]

PUB setScreenChar(x, y, c) | p
'' Set the character at the absolute screen coordinates x,y to c
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return
  byte[y * cols + x + long[ldr#screenPtr]] := c

PUB windowChar(x, y) | p
'' Return the character at the window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+vga_wcols] or y => byte[p+vga_wrows]
    return " "
  x += byte[p+vga_wleft]
  y += byte[p+vga_wtop]
  return byte[y * cols + x + long[ldr#screenPtr]]

PUB windowAtEnd(x, y) | p
'' Return true if window relative coordinates are past boundary
  p := long[ldr#videoPtr]
  return x => byte[p+vga_wcols] or y => byte[p+vga_wrows]

PUB windowCol | p
'' Return the current window relative x coordinate (column)
  p := long[ldr#videoPtr]
  return byte[p+vga_col] - byte[p+vga_wleft]

PUB windowRow | p
'' Return the current window relative y coordinate (row)
  p := long[ldr#videoPtr]
  return byte[p+vga_row] - byte[p+vga_wtop]

PUB out2(a, b)
'' Output a 2 character sequence
  out(a)
  out(b)

PUB out3(a, b, c)
'' Output a 3 character sequence
  out(a)
  out(b)
  out(c)

PUB out4(a, b, c, d)
'' Output a 4 character sequence
  out(a)
  out(b)
  out(c)
  out(d)

PUB out5(a, b, c, d, e)
'' Output a 5 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)

PUB out6(a, b, c, d, e, f)
'' Output a 6 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)
  out(f)

PRI print(c) | p
  p := long[ldr#videoPtr]
  if byte[p+vga_eol] <> 0
    newline
  byte[long[ldr#screenPtr] + byte[p+vga_row] * cols + byte[p+vga_col]] := c
  updateCursor(+1,0,relX|relY,true)

PRI updateCursor(x,y,mode,chkEol) | p, ax, ay, wx, wy, m, q, i
  p := long[ldr#videoPtr]
  ax := byte[p+vga_wleft]
  ay := byte[p+vga_wtop]
  wx := ax + byte[p+vga_wcols] - 1
  wy := ay + byte[p+vga_wrows] - 1
  case mode & %0011
    relX:
      x += byte[p+vga_col]
      x &= x & $80000000 == 0
    winX:
      x += ax
  case (mode & %1100) >> 2
    relY >> 2:
      y += byte[p+vga_row]
      y &= y & $80000000 == 0
    winY >> 2:
      y += ay
  if chkEol
    byte[p+vga_eol] := x > wx or y > wy
  byte[p+vga_col] := x := x #> ax <# wx
  byte[p+vga_row] := y := y #> ay <# wy
  if m := byte[p+vga_mode]
    q := long[ldr#cursorPtr]
    repeat i from 0 to 1
      if updateMask << i & m
        byte[3*i+q][0] := x
        byte[3*i+q][1] := y

PRI newline | p
  p := long[ldr#videoPtr]
  updateCursor(0,+1,winX|relY,true)
  if byte[p+vga_eol] <> 0
    scrollUpWin
  byte[p+vga_eol]~

PRI scrollUpWin | i, k, p, s, e
  p := long[ldr#videoPtr]
  s := long[ldr#screenPtr]
  if byte[p+vga_wrows] > 1
    e := byte[p+vga_wtop] + byte[p+vga_wrows] - 1
    repeat i from byte[p+vga_wtop] to e - 1
      moveline(byte[p+vga_wleft],i,byte[p+vga_wcols],i+1)
    k := byte[p+vga_col]
    updateCursor(0,e,winX|absY,false)
    cleartoeol
    updateCursor(k,0,absX|relY,false)

PRI scrollDownWin | i, k, p, s, e
  p := long[ldr#videoPtr]
  s := long[ldr#screenPtr]
  if byte[p+vga_wrows] > 1
    e := byte[p+vga_wtop] + byte[p+vga_wrows] - 1
    repeat i from e to byte[p+vga_wtop] + 1
      moveline(byte[p+vga_wleft],i,byte[p+vga_wcols],i-1)
    k := byte[p+vga_col]
    updateCursor(0,0,winX|winY,false)
    cleartoeol
    updateCursor(k,0,absX|winY,false)

PRI cleartoeol | p, x, w
 p := long[ldr#videoPtr]
 x := byte[p+vga_col]
 w := byte[p+vga_wleft] + byte[p+vga_wcols] - 1
 fillwith(x,byte[p+vga_row],w-x," ")
       
PRI fillwith(x, y, w, c) | i, s
  s := long[ldr#screenPtr] + y * cols
  repeat i from x to x + w - 1
    byte[s + i] := c

PRI moveline(x, y, w, f) | i, s, t
  s := long[ldr#screenPtr] + y * cols
  t := long[ldr#screenPtr] + f * cols
  repeat i from x to x + w - 1
    byte[s + i] := byte[t + i]

PUB debugOut(x, y, value)
  setScreenChar(x++,y,value)
  return x

PUB debugHex(x, y, value, digits)
'' Print a hexadecimal number
  value <<= (8 - digits) << 2
  repeat digits
    setScreenChar(x++,y,lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
  return x

PUB debugBin(x, y, value, digits)
'' Print a binary number
  value <<= 32 - digits
  repeat digits
    setScreenChar(x++,y,(value <-= 1) & 1 + "0")
  return x

PUB debugStr(x, y, stringptr)
'' Print a zero-terminated string
  repeat strsize(stringptr)
    setScreenChar(x++,y,byte[stringptr++])
  return x

PUB debugDec(x,y,value) | i
'' Print a decimal number
  if value < 0
    -value
    setScreenChar(x++,y,"-")
  i := 1_000_000_000
  repeat 10
    if value => i
      setScreenChar(x++,y,value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      setScreenChar(x++,y,"0")
    i /= 10
  return x

PUB debugDecFixed(x, y, value, width) | i, p, f, a, b, c
'' Print a decimal number of fixed width
  p := @a
  f~
  if value < 0
    -value
    byte[p++] := "-"
  i := 1_000_000_000
  repeat 10
    if value => i
      byte[p++] := value / i + "0"
      value //= i
      f~~
    elseif f or i == 1
      byte[p++] := "0"
    i /= 10
  byte[p]~
  i := strsize(@a)
  if i < width
    repeat width - i
      setScreenChar(x++,y," ")
  return debugStr(x,y,@a)

PUB debugChr(x, y, value, width)
'' Print up to 4 characters LSB first
  repeat width <# 4
    setScreenChar(x++,y,value)
    value >>= 8 